#include <cmath>
#include <cstddef>
#include <cstdlib>
#include <iostream>
#include <limits>
#include <ostream>
#include <random>

#include <cublas_v2.h>

#include <thrust/copy.h>
#include <thrust/device_vector.h>
#include <thrust/host_vector.h>

#include "timer.h"
using namespace visionaray;


//-------------------------------------------------------------------------------------------------
// Calculate dense matrix-matrix product C=AB, where (rows x cols) A is N x M and B is M x N
// Unoptimized, i.e. does not use shared memory
//

__global__ void matrix_mul_unoptimized(float* C, float* A, float* B, int N, int M)
{
}


//-------------------------------------------------------------------------------------------------
// Calculate dense matrix-matrix product C=AB, where (rows x cols) A is N x M and B is M x N
// Use CUDA shared memory
//

template <int BS = 32> // blocks of size BS x BX
__global__ void matrix_mul_shared_mem(float* C, float* A, float* B, int N, int M)
{
}


//-------------------------------------------------------------------------------------------------
// Calculate dense matrix-matrix product C=AB, where (rows x cols) A is N x M and B is M x N
// Use cublas
//

void matrix_mul_cublas(float* d_C, float* d_A, float* d_B, int N, int M)
{
    cublasHandle_t handle;
    cublasCreate(&handle);

    float alpha = 1.0f;
    float beta  = 0.0f;

    // CUBLAS MM is column-major, thus reverse input order
    cublasSgemm(
        handle,
        CUBLAS_OP_N,
        CUBLAS_OP_N,
        N,
        N,
        M,
        &alpha,
        d_B,
        N,
        d_A,
        M,
        &beta,
        d_C,
        N
        );
}


int main()
{
    // For simplicity assume that N and M are multiples of 32
    constexpr static int N = 2048;
    constexpr static int M = 1536;

    // Make some random matrices
    thrust::host_vector<float> h_A(N * M);
    thrust::host_vector<float> h_B(M * N);

    std::default_random_engine rng;
    std::uniform_real_distribution<float> dist;

    for (std::size_t i = 0; i < N * M; ++i)
    {
        h_A[i] = dist(rng);
        h_B[i] = dist(rng);
    }

    // Copy to GPU
    thrust::device_vector<float> d_C(N * N);
    thrust::device_vector<float> d_A(h_A);
    thrust::device_vector<float> d_B(h_B);

    float* C = thrust::raw_pointer_cast(d_C.data());
    float* A = thrust::raw_pointer_cast(d_A.data());
    float* B = thrust::raw_pointer_cast(d_B.data());

    // divide and round up
    auto div_up = [](unsigned a, unsigned b)
    {   
        return (a + b - 1) / b;
    };

    // Calculate grid dimensions
    dim3 threads_per_block(32, 32);
    dim3 num_blocks(
        div_up(N, threads_per_block.x),
        div_up(N, threads_per_block.y)
        );


    // Before we measure execution time, perform
    // some warmup runs with each method to fill
    // the GPU caches!
    constexpr static int warmup = 20;


    // Matrix multiplication w/o shared memory

    for (int i = 0; i < warmup; ++i)
        matrix_mul_unoptimized<<<num_blocks, threads_per_block>>>(C, A, B, N, M);

    cuda::timer t;
    matrix_mul_unoptimized<<<num_blocks, threads_per_block>>>(C, A, B, N, M);
    std::cout << "Elapsed time (w/o shared mem): " << t.elapsed() << " ms.\n";
    thrust::host_vector<float> h_C_unoptimized(d_C);


    // Matrix multiplication w/ shared memory

    for (int i = 0; i < warmup; ++i)
        matrix_mul_shared_mem<<<num_blocks, threads_per_block>>>(C, A, B, N, M);

    t.reset();
    matrix_mul_shared_mem<<<num_blocks, threads_per_block>>>(C, A, B, N, M);
    std::cout << "Elapsed time (w/ shared mem):  " << t.elapsed() << " ms.\n";
    thrust::host_vector<float> h_C_shared_mem(d_C);

    // Matrix multiplication with cublas (for verification)

    for (int i = 0; i < warmup; ++i)
        matrix_mul_cublas(C, A, B, N, M);

    t.reset();
    matrix_mul_cublas(C, A, B, N, M);
    std::cout << "Elapsed time (cublasSgemm):    " << t.elapsed() << " ms.\n";
    thrust::host_vector<float> h_C_cublas(d_C);


    // Verify our results with cublas
    // (Use a rather large epsilon, cublas optimizes quite severely
    // and trades speed for accuracy)
    for (std::size_t i = 0; i < N * N; ++i)
    {
        float diff = std::abs(h_C_unoptimized[i] - h_C_cublas[i]);
        if (diff > 1E-2)
        {
            std::cerr << "Incorrect result (unoptimized routine)\n";
            return EXIT_FAILURE;
        }

        diff = std::abs(h_C_shared_mem[i] - h_C_cublas[i]);
        if (diff > 1E-2)
        {
            std::cerr << "Incorrect result (shared mem routine)\n";
            return EXIT_FAILURE;
        }
    }

    return EXIT_SUCCESS;
}
